;; Generated package description from leaf.el  -*- no-byte-compile: t -*-
(define-package "leaf" "4.4.8.0.20210601.162512" "Simplify your init.el configuration, extended use-package" '((emacs "24.1")) :keywords '("lisp" "settings") :authors '(("Naoya Yamashita" . "conao3@gmail.com")) :maintainer '("Naoya Yamashita" . "conao3@gmail.com") :url "https://github.com/conao3/leaf.el")
